from piccolo import columns as col
from piccolo.engine.postgres import PostgresEngine

import settings
from apphelpers.db.piccolo import BaseTable

db = PostgresEngine(
    config=dict(
        host=settings.DB_HOST,
        database=settings.DB_NAME,
        user=settings.DB_USER,
        password=settings.DB_PASS,
    )
)


class Book(BaseTable, db=db):
    name = col.Text()



from unittest import mock
import httpx

import pytest
from requests.exceptions import HTTPError

import apphelpers.sessions as sessionslib
from apphelpers.errors.fastapi import BaseError
from apphelpers.rest.fastapi import honeybadger_wrapper

base_url = "http://127.0.0.1:5000/"
echo_url = base_url + "echo"
echo_header_url = base_url + "echo-header"
echo_async_url = base_url + "echo-async"
secure_echo_url = base_url + "secure-echo"
echo_groups_url = base_url + "echo-groups"
echo_groups_async_url = base_url + "echo-groups-async"
echo_site_groups_url = base_url + "sites/{site_id}/echo-groups"
echo_site_groups_async_url = base_url + "sites/{site_id}/echo-groups-async"
pid_path = "tests/run/app.pid"


@pytest.mark.asyncio
class TestRest:

    async def test_get(self, client: httpx.AsyncClient):
        word = "hello"
        url = echo_url + "/" + word
        response = await client.get(url, params=dict())
        assert response.json() == word

        word = "hello"
        url = echo_url + "/" + word
        params = {"word": word}
        response = await client.get(url, params=params)
        assert response.json() == word

    async def test_get_async(self, client: httpx.AsyncClient):
        word = "hello"
        url = echo_async_url + "/" + word
        response = await client.get(url, params=dict())
        assert response.json() == word

        word = "hello"
        url = echo_async_url + "/" + word
        params = {"word": word}
        response = await client.get(url, params=params)
        assert response.json() == word

    async def test_get_header(self, client: httpx.AsyncClient):
        x_key = "secret key"
        url = echo_header_url
        headers = {"X-KEY": x_key}
        response = await client.get(url, headers=headers)
        assert response.json() == x_key

    async def test_get_multi_params(self, client: httpx.AsyncClient):
        nums = [3, 5, 8]
        url = base_url + "add"
        params = {"nums": nums}
        response = await client.get(url, params=params)
        assert response.json() == 8  # sum(nums)

    async def test_post(self, client: httpx.AsyncClient):
        word = "hello"
        url = echo_url
        response = await client.post(url, json={"word": word})
        assert response.json()["word"] == word

    async def test_secure_echo(
        self, client: httpx.AsyncClient, sessionsdb: sessionslib.SessionDBHandler
    ):
        word = "hello"
        headers = {"NoAuthorization": "Header"}
        url = secure_echo_url + "/" + word
        resp = await client.get(url, headers=headers)
        assert resp.status_code == 401

        uid = 10000
        d = dict(uid=uid, groups=[])
        sid = await sessionsdb.create(**d)

        headers = {"Authorization": sid}
        resp = await client.get(url, headers=headers)
        assert resp.status_code == 200
        assert resp.json() == ("%s:%s" % (uid, word))

        client.cookies.set("__s", sid)
        resp = await client.get(url)
        client.cookies.delete("__s")
        assert resp.status_code == 200
        assert resp.json() == ("%s:%s" % (uid, word))

    async def test_user_id(
        self, client: httpx.AsyncClient, sessionsdb: sessionslib.SessionDBHandler
    ):
        uid = 101
        d = dict(uid=uid, groups=[])
        sid = await sessionsdb.create(**d)

        headers = {"Authorization": sid}

        word = "hello"
        url = echo_url + "/" + word
        resposne = await client.get(url, headers=headers)
        assert resposne.json() == ("%s:%s" % (uid, word))

        url = base_url + "me/uid"

        data = {"uid": None}
        resp = await client.post(url, json=data, headers=headers)
        assert resp.json() == data["uid"]

        data = {"uid": 1}
        resp = await client.post(url, json=data, headers=headers)
        assert resp.json() == data["uid"]

        data = {"uid": uid}
        resp = await client.post(url, json=data, headers=headers)
        assert resp.json() == data["uid"]

        url = base_url + "me/uid-unpacked"

        data = {"uid": None}
        resp = await client.post(url, json=data, headers=headers)
        assert resp.json() == data["uid"]

        data = {"uid": 1}
        resp = await client.post(url, json=data, headers=headers)
        assert resp.json() == data["uid"]

        data = {"uid": uid}
        resp = await client.post(url, json=data, headers=headers)
        assert resp.json() == data["uid"]

    async def test_group_access(
        self, client: httpx.AsyncClient, sessionsdb: sessionslib.SessionDBHandler
    ):
        url = echo_groups_url

        # 1. No group
        uid = 111
        groups = []
        d = dict(uid=uid, groups=groups)
        sid = await sessionsdb.create(**d)

        headers = {"Authorization": sid}
        response = await client.get(url, headers=headers)
        assert response.status_code == 403

        # 2. Forbidden group
        uid = 112
        groups = ["noaccess-group"]
        d = dict(uid=uid, groups=groups)
        sid = await sessionsdb.create(**d)

        headers = {"Authorization": sid}
        response = await client.get(url, headers=headers)
        assert response.status_code == 403

        # 3. Access group
        uid = 113
        groups = ["access-group"]
        d = dict(uid=uid, groups=groups)
        sid = await sessionsdb.create(**d)

        headers = {"Authorization": sid}
        response = await client.get(url, headers=headers)
        assert response.status_code == 200
        assert response.json() == groups

    async def test_group_access_async(
        self, client: httpx.AsyncClient, sessionsdb: sessionslib.SessionDBHandler
    ):
        url = echo_groups_async_url

        # 1. No group
        uid = 111
        groups = []
        d = dict(uid=uid, groups=groups)
        sid = await sessionsdb.create(**d)

        headers = {"Authorization": sid}
        response = await client.get(url, headers=headers)
        assert response.status_code == 403

        # 2. Forbidden group
        uid = 112
        groups = ["noaccess-group"]
        d = dict(uid=uid, groups=groups)
        sid = await sessionsdb.create(**d)

        headers = {"Authorization": sid}
        response = await client.get(url, headers=headers)
        assert response.status_code == 403

        # 3. Access group
        uid = 113
        groups = ["access-group"]
        d = dict(uid=uid, groups=groups)
        sid = await sessionsdb.create(**d)

        headers = {"Authorization": sid}
        response = await client.get(url, headers=headers)
        assert response.status_code == 200
        assert response.json() == groups

    async def test_not_found_on_none(self, client: httpx.AsyncClient):
        url = base_url + "snakes/viper"
        response = await client.get(url)
        assert response.status_code != 404

        url = base_url + "snakes"
        response = await client.get(url)
        assert response.status_code == 404

    async def test_get_fields(self, client: httpx.AsyncClient):
        url = base_url + "fields"
        response = await client.get(url)
        assert response.json() == {}

        url = base_url + "fields?fields=foo"
        response = await client.get(url)
        assert response.json() == {"foo": 1}

        url = base_url + "fields?fields=bar"
        response = await client.get(url)
        assert response.json() == {"bar": None}

        url = base_url + "fields?fields=foo&fields=bar"
        response = await client.get(url)
        assert response.json() == {"foo": 1, "bar": None}

    async def test_site_group_access(
        self, client: httpx.AsyncClient, sessionsdb: sessionslib.SessionDBHandler
    ):
        url = echo_site_groups_url

        # 1. No group
        uid = 1111
        site_id = 2011
        site_groups = {2011: []}
        d = dict(uid=uid, site_groups=site_groups)
        sid = await sessionsdb.create(**d)

        headers = {"Authorization": sid}
        response = await client.get(url.format(site_id=site_id), headers=headers)
        assert response.status_code == 403

        # 2. Forbidden group
        uid = 1112
        site_groups = {2011: ["noaccess-group"]}
        d = dict(uid=uid, site_groups=site_groups)
        sid = await sessionsdb.create(**d)

        headers = {"Authorization": sid}
        response = await client.get(url.format(site_id=site_id), headers=headers)
        assert response.status_code == 403

        # 3. Access group
        uid = 1113
        site_groups = {2011: ["access-group"]}
        d = dict(uid=uid, site_groups=site_groups)
        sid = await sessionsdb.create(**d)

        headers = {"Authorization": sid}
        response = await client.get(url.format(site_id=site_id), headers=headers)
        assert response.status_code == 200
        assert response.json() == site_groups[site_id]

    async def test_site_group_access_async(
        self, client: httpx.AsyncClient, sessionsdb: sessionslib.SessionDBHandler
    ):
        url = echo_site_groups_async_url

        # 1. No group
        uid = 1211
        site_id = 3011
        site_groups = {3011: []}
        d = dict(uid=uid, site_groups=site_groups)
        sid = await sessionsdb.create(**d)

        headers = {"Authorization": sid}
        response = await client.get(url.format(site_id=site_id), headers=headers)
        assert response.status_code == 403

        # 2. Forbidden group
        uid = 1212
        site_groups = {3011: ["noaccess-group"]}
        d = dict(uid=uid, site_groups=site_groups)
        sid = await sessionsdb.create(**d)

        headers = {"Authorization": sid}
        response = await client.get(url.format(site_id=site_id), headers=headers)
        assert response.status_code == 403

        # 3. Access group
        uid = 1213
        site_groups = {3011: ["access-group"]}
        d = dict(uid=uid, site_groups=site_groups)
        sid = await sessionsdb.create(**d)

        headers = {"Authorization": sid}
        response = await client.get(url.format(site_id=site_id), headers=headers)
        assert response.status_code == 200
        assert response.json() == site_groups[site_id]

    async def test_not_found_on_none_async(self, client: httpx.AsyncClient):
        url = base_url + "snakes-async/viper"
        response = await client.get(url)
        assert response.status_code != 404

        url = base_url + "snakes-async"
        response = await client.get(url)
        assert response.status_code == 404

    async def test_user_agent_async_and_site_ctx(
        self, client: httpx.AsyncClient, sessionsdb: sessionslib.SessionDBHandler
    ):
        url = base_url + "echo-user-agent-async"

        headers = {"Authorization": await sessionsdb.create(uid=1214)}
        response = await client.get(url, headers=headers)
        assert response.status_code == 200
        assert "python-httpx" in response.text

        headers = {"Authorization": await sessionsdb.create(uid=1215, site_ctx=4011)}
        response = await client.get(url, headers=headers)
        assert response.status_code == 401

        url = base_url + "echo-user-agent-without-site-ctx-async"

        headers = {"Authorization": await sessionsdb.create(uid=1214)}
        response = await client.get(url, headers=headers)
        assert response.status_code == 200
        assert "python-httpx" in response.text

        headers = {"Authorization": await sessionsdb.create(uid=1215, site_ctx=4011)}
        response = await client.get(url, headers=headers)
        assert response.status_code == 200
        assert "python-httpx" in response.text

    async def test_piccolo(self, client: httpx.AsyncClient):
        url = base_url + "count-books"
        response = await client.get(url)
        assert response.json() == 0

        url = base_url + "add-books"
        data = {"succeed": True}
        response = await client.post(url, params=data)
        assert response.status_code == 200

        url = base_url + "count-books"
        response = await client.get(url)
        assert response.json() == 3

        url = base_url + "count-books"
        response = await client.get(url)
        assert response.json() == 3

    async def test_honeybadger_wrapper(self):

        mocked_honeybadger = mock.MagicMock()
        wrapper = honeybadger_wrapper(mocked_honeybadger)

        def good_endpoint(foo):
            return foo

        class IgnorableError(BaseError):
            report = False

        async def bad_endpoint(foo):
            raise IgnorableError()

        async def worse_endpoint(foo, password):
            raise BaseError()

        async def worst_endpoint(foo):
            raise RuntimeError()

        wrapped_good_endpoint = wrapper(good_endpoint)
        wrapped_bad_endpoint = wrapper(bad_endpoint)
        wrapped_worse_endpoint = wrapper(worse_endpoint)
        wrapped_worst_endpoint = wrapper(worst_endpoint)

        assert wrapped_good_endpoint(1) == 1
        assert not mocked_honeybadger.notify.called

        with pytest.raises(IgnorableError):
            await wrapped_bad_endpoint(1)
        assert not mocked_honeybadger.notify.called

        with pytest.raises(BaseError) as e:
            await wrapped_worse_endpoint(1, password="secret")
        mocked_honeybadger.notify.assert_called_once_with(
            e.value,
            context={
                "func": "worse_endpoint",
                "args": (1,),
                "kwargs": {"password": "[FILTERED]"},
            },
        )
        assert mocked_honeybadger.notify.call_count == 1

        with pytest.raises(RuntimeError):
            await wrapped_worst_endpoint(1)
        assert mocked_honeybadger.notify.call_count == 2

        mocked_honeybadger.notify.side_effect = HTTPError(
            response=mock.MagicMock(status_code=403)
        )
        with pytest.raises(RuntimeError):
            await wrapped_worst_endpoint(1)
        # TODO: How to check nested exception?
        assert mocked_honeybadger.notify.call_count == 3

        mocked_honeybadger.notify.side_effect = HTTPError(
            response=mock.MagicMock(status_code=401)
        )
        with pytest.raises(HTTPError):
            await wrapped_worst_endpoint(1)
        # TODO: How to check nested exception?
        assert mocked_honeybadger.notify.call_count == 4



from __future__ import annotations

from collections.abc import Iterable
from enum import Enum


def _unpack(objs):
    if isinstance(objs, Iterable) and not isinstance(objs, (str, bytes)):
        for obj in objs:
            yield from _unpack(obj)
    elif isinstance(objs, Enum):
        yield objs.value
    else:
        yield objs


def login_required(func):
    """Auth token is required for this endpoint."""

    func.login_required = True
    return func


def login_optional(func):
    """Auth token is optional for this endpoint."""

    func.login_optional = True
    return func


def skip_authorization(func):
    """Auth token is not required for this endpoint."""

    func.skip_authorization = True
    return func


def auth_by_cookie_or_header(func):
    """Auth token can be provided by cookie or header."""

    func.auth_by_cookie_or_header = True
    return func


def any_group_required(*groups):
    """Any of the mentioned groups is required."""

    def decorator(func):
        func.any_group_required = set(_unpack(groups))
        return login_required(func)

    return decorator


def all_groups_required(*groups):
    """All of the mentioned groups are required."""

    def decorator(func):
        func.all_groups_required = set(_unpack(groups))
        return login_required(func)

    return decorator


def groups_forbidden(*groups):
    """None of the mentioned groups are allowed."""

    def decorator(func):
        func.groups_forbidden = set(_unpack(groups))
        return login_required(func)

    return decorator


def authorizer(authorizer):
    """Custom the authorizer for this endpoint."""

    def decorator(func):
        func.authorizer = authorizer
        return login_required(func)

    return decorator


def ignore_site_ctx(func):
    """Ignore site context for this endpoint."""

    func.ignore_site_ctx = True
    return login_required(func)


def not_found_on_none(func):
    """If the return value is None, return HTTP 404 Not Found error."""

    func.not_found_on_none = True
    return func


def response_model(response_model):
    """[FastAPI only] Set response model for this endpoint."""

    def decorator(func):
        func.response_model = response_model
        return func

    return decorator


def skip_dbtransaction(func):
    """Skip database transaction for this endpoint."""

    func.skip_dbtransaction = True
    return func



"""
This file marks the root of tests.
Common pytest fixtures can be defines here.
See https://docs.pytest.org/en/latest/reference/fixtures.html
"""

import asyncio
import pytest

import settings
from apphelpers.async_sessions import SessionDBHandler


sessiondb_conn = dict(
    host=settings.SESSIONSDB_HOST,
    port=settings.SESSIONSDB_PORT,
    password=settings.SESSIONSDB_PASSWD,
    db=settings.SESSIONSDB_NO,
)


@pytest.fixture(scope="module")
def anyio_backend():
    return "asyncio"


@pytest.fixture
def event_loop():
    loop = asyncio.get_event_loop()
    yield loop
    loop.close()


@pytest.fixture
async def sessionsdb():
    sessionsdb = SessionDBHandler(sessiondb_conn)
    yield sessionsdb
    await sessionsdb.close()



from typing import Optional

from fastapi import Query
from pydantic import BaseModel

from apphelpers.rest import endpoint as ep
from apphelpers.rest.fastapi import RequestHeaders, json_body, AuthParams, BodyParams
from fastapi_tests.app.models import Book


@ep.login_optional
def echo(word, user: AuthParams.user):
    return "%s:%s" % (user.id, word) if user else word


@ep.login_optional
async def echo_async(word, user: AuthParams.user):
    return "%s:%s" % (user.id, word) if user else word


@ep.login_optional
def echo_post(data: json_body, user: AuthParams.user):
    return "%s:%s" % (user.id, data) if user else data


@ep.skip_authorization
def echo_header(x_key: RequestHeaders.custom):
    return x_key


@ep.login_required
@ep.auth_by_cookie_or_header
def secure_echo(word, user_id: AuthParams.user_id):
    return "%s:%s" % (user_id, word) if user_id else word


@ep.groups_forbidden("noaccess-group")
@ep.all_groups_required("access-group")
def echo_groups(user: AuthParams.user):
    return user.groups


@ep.groups_forbidden("noaccess-group")
@ep.all_groups_required("access-group")
async def echo_groups_async(user: AuthParams.user):
    return user.groups


@ep.skip_authorization
def add(nums):
    return sum(int(x) for x in nums)


@ep.login_required
def get_my_uid(body: json_body):
    return body["uid"]


@ep.login_required
def get_my_uid_unpacked(uid: BodyParams.int = None):
    return uid


@ep.login_required
@ep.response_model(str)
@ep.not_found_on_none
def get_snake(name=None):
    return name


@ep.login_required
@ep.response_model(str)
@ep.not_found_on_none
async def get_snake_async(name=None):
    return name


@ep.all_groups_required("access-group")
@ep.groups_forbidden("noaccess-group")
def echo_site_groups(site_id: int, user: AuthParams.user):
    return user.site_groups[site_id]


@ep.all_groups_required("access-group")
@ep.groups_forbidden("noaccess-group")
async def echo_site_groups_async(site_id: int, user: AuthParams.user):
    return user.site_groups[site_id]


@ep.login_required
async def echo_user_agent_async(user_agent: RequestHeaders.user_agent):
    return user_agent


@ep.login_required
@ep.ignore_site_ctx
async def echo_user_agent_without_site_ctx_async(user_agent: RequestHeaders.user_agent):
    return user_agent


class Fields(BaseModel):
    foo: Optional[int] = None
    bar: Optional[int] = None


@ep.response_model(Fields)
@ep.skip_authorization
async def get_fields(fields: set = Query(..., default_factory=set)):
    data = {"foo": 1, "bar": None}
    return {k: v for k, v in data.items() if k in fields}


@ep.skip_authorization
async def add_books(succeed: bool):
    await Book.insert(Book(name="The Pillars of the Earth")).run()
    await Book.insert(Book(name="The Cathedral and the Bazaar")).run()
    if not succeed:
        raise ValueError("Failure")
    await Book.insert(Book(name="The Ego Trick")).run()


@ep.skip_authorization
async def count_books():
    return await Book.count()


def setup_routes(factory):
    factory.get("/echo/{word}")(echo)
    factory.get("/echo-async/{word}")(echo_async)
    factory.post("/echo")(echo_post)
    factory.get("/echo-header")(echo_header)

    factory.get("/add")(add)

    factory.get("/secure-echo/{word}")(secure_echo)
    factory.get("/echo-groups")(echo_groups)
    factory.get("/echo-groups-async")(echo_groups_async)

    factory.post("/me/uid")(get_my_uid)
    factory.post("/me/uid-unpacked")(get_my_uid_unpacked)

    factory.get("/snakes/{name}")(get_snake)
    factory.get("/snakes-async/{name}")(get_snake_async)

    factory.get("/sites/{site_id}/echo-groups")(echo_site_groups)
    factory.get("/sites/{site_id}/echo-groups-async")(echo_site_groups_async)

    factory.get("/echo-user-agent-async")(echo_user_agent_async)
    factory.get("/echo-user-agent-without-site-ctx-async")(
        echo_user_agent_without_site_ctx_async
    )
    factory.get("/fields")(get_fields)
    factory.get("/count-books")(count_books)
    factory.post("/add-books")(add_books)



from __future__ import annotations

import copy
from dataclasses import asdict, dataclass, field
from typing import Dict, List, Optional

from converge import settings
from requests.exceptions import HTTPError

if settings.get("HONEYBADGER_API_KEY"):
    from honeybadger.utils import filter_dict


def phony(f):
    return f


@dataclass
class User:
    sid: Optional[str] = None
    id: Optional[int] = None
    name: Optional[str] = None
    groups: List[str] = field(default_factory=list)
    email: Optional[str] = None
    mobile: Optional[str] = None
    site_groups: Dict[int, int] = field(default_factory=dict)
    site_ctx: Optional[int] = None

    def to_dict(self):
        return asdict(self)

    def __bool__(self):
        return self.id is not None


def notify_honeybadger(honeybadger, error, func, args, kwargs):
    try:
        honeybadger.notify(
            error,
            context={
                "func": func.__name__,
                "args": args,
                "kwargs": filter_dict(copy.deepcopy(kwargs), settings.HB_PARAM_FILTERS),
            },
        )
    except HTTPError as e:
        if e.response.status_code == 403:
            # Ignore 403 Forbidden errors. We get alerted by HB anyway.
            pass
        else:
            raise e



# Test starting (initializing) apphelpers app with different supported configurations


def test_basic():
    import tests.services.basic  # noqa: F401


def test_session():
    import tests.services.session  # noqa: F401



from __future__ import annotations

import json
from typing import Any, ClassVar, List, Optional

from redis import Redis


class ReadOnlyCachedModel:
    """
    Read only cached model
    """

    connection: ClassVar[Redis]
    ns: ClassVar[str]
    key_fields: ClassVar[List[str]]
    secondary_key_fields: ClassVar[List[str]]

    @classmethod
    def _prefix_key(cls, data: dict) -> str:
        key = cls.ns
        for _field in cls.key_fields:
            key += f":{data[_field]}"
        return key

    @classmethod
    def _secondary_prefix_key(cls, data: dict) -> str:
        key = f"{cls.ns}:_sk_"
        for _field in cls.secondary_key_fields:
            key += f":{data.get(_field, '*')}"
        return key

    @classmethod
    def _get_matched_keys(cls, data: dict) -> List[str]:
        pattern = cls.ns
        for _field in cls.key_fields:
            pattern += f':{data.get(_field, "*")}'
        return cls.connection.keys(pattern)  # type: ignore

    @classmethod
    def get(cls, **data: Any) -> Optional[dict]:
        key = cls._prefix_key(data)
        value: Any = cls.connection.get(key)
        return json.loads(value) if value else None

    @classmethod
    def get_by_secondary_key(cls, **data: Any) -> Optional[dict]:
        secondary_key = cls._secondary_prefix_key(data)
        primary_key = cls.connection.get(secondary_key)
        if primary_key:
            value = cls.connection.get(primary_key)
            return json.loads(value) if value else None

    @classmethod
    def exists(cls, **data: Any) -> bool:
        key = cls._prefix_key(data)
        return cls.connection.exists(key)  # type: ignore

    @classmethod
    def get_count(cls, **data: Any) -> int:
        key: Any = cls._prefix_key(data)
        count: Optional[Any] = cls.connection.get(key)
        return int(count) if count else 0

    @classmethod
    def count_matched_keys(cls, **data: Any) -> int:
        keys = cls._get_matched_keys(data)
        return len(keys)


class ReadWriteCachedModel(ReadOnlyCachedModel):
    """
    Read/Write cached model
    """

    timeout: ClassVar[Optional[int]] = None

    @classmethod
    def create(cls, **data: Any) -> str:
        key = cls._prefix_key(data)
        cls.connection.set(key, json.dumps(data))
        if cls.timeout:
            cls.connection.expire(key, cls.timeout)
        return key

    @classmethod
    def add_secondary_key(cls, primary_key: str, **data: Any) -> str:
        secondary_key = cls._secondary_prefix_key(data)
        cls.connection.set(secondary_key, primary_key)
        return secondary_key

    @classmethod
    def create_lookup(cls, **data: Any) -> str:
        key = cls._prefix_key(data)
        cls.connection.set(key, 1)
        if cls.timeout:
            cls.connection.expire(key, cls.timeout)
        return key

    @classmethod
    def create_counter(cls, starting=1, **data: Any) -> str:
        key = cls._prefix_key(data)
        cls.connection.set(key, starting)
        if cls.timeout:
            cls.connection.expire(key, cls.timeout)
        return key

    @classmethod
    def update(cls, **data):
        # key = cls._prefix_key(data)
        # NOTE: keepttl works only with Redis 6.0+
        # cls.connection.set(key, json.dumps(data), keepttl=True)
        cls.create(**data)

    @classmethod
    def increment(cls, amount=1, **data):
        key = cls._prefix_key(data)
        cls.connection.incr(key, amount)

    @classmethod
    def decrement(cls, amount=1, **data):
        key = cls._prefix_key(data)
        cls.connection.decr(key, amount)

    @classmethod
    def delete(cls, **data):
        key = cls._prefix_key(data)
        cls.connection.delete(key)

    @classmethod
    def delete_secondary_key(cls, **data: Any):
        secondary_key = cls._secondary_prefix_key(data)
        cls.connection.delete(secondary_key)

    @classmethod
    def delete_all(cls, **data):
        keys = cls._get_matched_keys(data)
        if keys:
            cls.connection.delete(*keys)

    @classmethod
    def delete_all_secondary_keys(cls):
        secondary_keys = cls.connection.keys(cls._secondary_prefix_key(data={}))
        if secondary_keys:
            cls.connection.delete(*secondary_keys)



from contextlib import asynccontextmanager
from functools import wraps
from typing import List, Set, Type

from piccolo.engine.postgres import PostgresEngine
from piccolo.table import Table, create_db_tables_sync, drop_db_tables_sync


@asynccontextmanager
async def connection_pool_lifespan(engine: PostgresEngine, **kwargs):
    print("db: starting connection pool")
    await engine.start_connection_pool(**kwargs)
    yield
    print("db: closing connection pool")
    await engine.close_connection_pool()


dbtransaction_ctx = PostgresEngine.transaction


def dbtransaction(engine: PostgresEngine, allow_nested=True):
    def wrapper(f):
        @wraps(f)
        async def f_wrapped(*args, **kw):
            async with dbtransaction_ctx(engine, allow_nested=allow_nested):
                return await f(*args, **kw)

        return f_wrapped

    return wrapper


class BaseTable(Table):
    @classmethod
    def all_column_names(cls) -> Set[str]:
        return {col._meta.name for col in cls._meta.columns}


def get_sub_tables(basetable: Type[Table]) -> List[Type[Table]]:
    tables = []
    for subtable in basetable.__subclasses__():
        tables.append(subtable)
        tables.extend(get_sub_tables(subtable))
    return tables


def setup_db(tables: List[Type[Table]]):
    create_db_tables_sync(*tables, if_not_exists=True)


def setup_db_from_basetable(basetable: Type[Table]):
    tables = get_sub_tables(basetable)
    setup_db(tables)


def destroy_db(tables: List[Type[Table]]):
    drop_db_tables_sync(*tables)


def destroy_db_from_basetable(basetable: Type[Table]):
    tables = get_sub_tables(basetable)
    destroy_db(tables)



try:
    from apphelpers.db.piccolo import *  # noqa: F401, F403

    peewee_enabled = False
    print("apphelpers: using Piccolo")

except ImportError:
    from apphelpers.db.peewee import *  # noqa: F401, F403

    peewee_enabled = True
    print("apphelpers: using Peewee")



import pytest
from peewee import TextField

import settings
from apphelpers.db.peewee import create_base_model, create_pgdb_pool, dbtransaction

db = create_pgdb_pool(
    host=settings.DB_HOST,
    database=settings.DB_NAME,
    user=settings.DB_USER,
    password=settings.DB_PASS,
)
BaseModel = create_base_model(db)
dbtransaction = dbtransaction(db)


class Book(BaseModel):
    name = TextField()


def _add_book(name):
    Book.create(name=name)


def _add_book_loser(name):
    _add_book(name)
    loser  # will raise  # noqa: F821


def setup_module():
    Book.create_table()


def teardown_module():
    Book.drop_table()


def test_add_with_tr():
    Book.delete().execute()

    add_book = dbtransaction(_add_book)
    name = "The Pillars of the Earth"
    add_book(name)
    names = [b.name for b in Book.select()]
    assert name in names

    add_book_loser = dbtransaction(_add_book_loser)
    name = "The Cathedral and the Bazaar"
    with pytest.raises(NameError):
        add_book_loser(name)
    names = [b.name for b in Book.select()]
    assert name not in names

    add_book = dbtransaction(_add_book)
    name = "The Ego Trick"
    add_book(name)
    names = [b.name for b in Book.select()]
    assert name in names



from collections import namedtuple

import pytest

import apphelpers.async_sessions as sessionslib
from apphelpers.errors import InvalidSessionError


Session = namedtuple("Session", ["uid", "groups", "k", "v"])


class data:
    session = Session(987651, ["admin", "member"], "foo", "bar")
    session_email = "noone@example.com"
    site_ctx = 123


@pytest.mark.anyio
class TestSessions:

    @pytest.fixture(autouse=True)
    async def setup_class(self, sessionsdb: sessionslib.SessionDBHandler):
        await sessionsdb.destroy_all()

    async def test_create(self, sessionsdb: sessionslib.SessionDBHandler):
        # test create
        d = dict(
            uid=data.session.uid,
            groups=data.session.groups,
            extras=dict(email=data.session_email),
        )
        sid = await sessionsdb.create(**d)
        assert len(sid) == 43
        sid_new = await sessionsdb.create(data.session.uid, data.session.groups)
        assert sid == sid_new == await sessionsdb.uid2sid(d["uid"])
        assert await sessionsdb.uid2bound_sids(d["uid"]) == []

        d = dict(
            uid=data.session.uid,
            groups=data.session.groups,
            extras=dict(email=data.session_email),
            site_ctx=data.site_ctx,
        )
        bound_sid = await sessionsdb.create(**d)
        assert bound_sid != sid
        assert bound_sid == await sessionsdb.uid2sid(d["uid"], data.site_ctx)
        assert await sessionsdb.uid2bound_sids(d["uid"]) == [bound_sid]
        assert await sessionsdb.exists(bound_sid)
        assert await sessionsdb.uid2bound_site_ids(d["uid"]) == [data.site_ctx]

        await sessionsdb.destroy_all_for_bound_site(data.site_ctx)
        assert not await sessionsdb.exists(bound_sid)

        d = dict(
            uid=data.session.uid,
            groups=data.session.groups,
            extras=dict(email=data.session_email),
            site_ctx=data.site_ctx,
        )
        bound_sid = await sessionsdb.create(**d)
        assert bound_sid != sid
        assert bound_sid == await sessionsdb.uid2sid(d["uid"], data.site_ctx)
        assert await sessionsdb.uid2bound_sids(d["uid"]) == [bound_sid]
        assert await sessionsdb.exists(bound_sid)

        await sessionsdb.destroy_bound_sessions_for(data.session.uid)
        assert not await sessionsdb.exists(bound_sid)

    async def test_update(self, sessionsdb: sessionslib.SessionDBHandler):
        d = dict(
            uid=data.session.uid,
            groups=data.session.groups,
            extras=dict(email=data.session_email),
        )
        sid = await sessionsdb.create(**d)

        k, v = data.session.k, data.session.v
        await sessionsdb.update(sid, {k: v})
        d = await sessionsdb.get(sid)
        assert d[k] == v
        assert d["email"] == data.session_email
        await sessionsdb.remove_from_session(sid, [k])
        d = await sessionsdb.get(sid)
        assert k not in d

    async def test_resync(self, sessionsdb: sessionslib.SessionDBHandler):
        d = dict(
            uid=data.session.uid,
            groups=data.session.groups,
            extras=dict(email=data.session_email),
        )
        sid = await sessionsdb.create(**d)

        k, v = data.session.k, data.session.v
        await sessionsdb.update(sid, {k: v})
        d = await sessionsdb.get(sid)
        assert d[k] == v

        d = dict(
            uid=data.session.uid,
            groups=data.session.groups,
            extras=dict(email=data.session_email),
        )
        await sessionsdb.resync(sid, d)
        d = await sessionsdb.get(sid)
        assert k not in d

        await sessionsdb.destroy(sid)
        with pytest.raises(InvalidSessionError):
            await sessionsdb.get(sid)

    async def test_session_lookup(self, sessionsdb: sessionslib.SessionDBHandler):
        uids = range(10000, 10010)
        groups = ["grp1", "grp2"]
        for uid in uids:
            sid = await sessionsdb.create(uid, groups)
            assert (await sessionsdb.sid2uidgroups(sid)) == (uid, groups)
            await sessionsdb.destroy(sid)
            with pytest.raises(InvalidSessionError):
                await sessionsdb.get(sid)



import json

import jwt
import requests

try:
    from converge import settings
except ImportError:
    import settings


def fetch_info(token):
    json_web_key_sets = requests.get("https://appleid.apple.com/auth/keys")
    public_keys = {}
    for jwk in json_web_key_sets.json()["keys"]:
        key_id = jwk["kid"]
        public_keys[key_id] = jwt.algorithms.RSAAlgorithm.from_jwk(json.dumps(jwk))

    key_id = jwt.get_unverified_header(token)["kid"]
    key = public_keys[key_id]
    payload = jwt.decode(
        token, key=key, audience=settings.APPLE_AUDIANCE, algorithms=["RS256"]
    )
    return payload



from unittest import mock

import falcon
import hug
import pytest
from converge import settings
from requests.exceptions import HTTPError

import apphelpers.sessions as sessionslib
from apphelpers.errors.hug import BaseError
from apphelpers.rest.hug import honeybadger_wrapper

from . import service
from .app.models import globalgroups, sitegroups


class urls:
    echo = "echo"
    echo_for_registered = "secure-echo"
    echo_for_groups = "echo-groups"
    echo_for_sitegroups = "sites/1/echo-groups"
    echo_for_all_groups = "sites/1/echo-all-groups"
    echo_for_custom_authorization = "custom-authorization-echo"


pid_path = "tests/run/app.pid"

sessiondb_conn = dict(
    host=settings.SESSIONSDB_HOST,
    port=settings.SESSIONSDB_PORT,
    password=settings.SESSIONSDB_PASSWD,
    db=settings.SESSIONSDB_NO,
)
sessionsdb = sessionslib.SessionDBHandler(sessiondb_conn)
sessionsdb.destroy_all()


def test_get():
    word = "hello"
    url = urls.echo + "/" + word
    assert hug.test.get(service, url).data == word


def test_get_params():
    word = "hello"
    url = urls.echo + "/" + word
    params = {"word": word}
    assert hug.test.get(service, url, params=params).data == word


def test_get_multi_params():
    nums = [3, 5]
    url = "add"
    params = {"nums": nums}
    assert hug.test.get(service, url, params=params).data == sum(nums)


def test_post():
    word = "hello"
    url = urls.echo
    assert hug.test.post(service, url, body={"word": word}).data == word


def test_echo_for_registered():
    word = "hello"
    headers = {"NoAuthorization": "Header"}
    url = urls.echo_for_registered + "/" + word
    resp = hug.test.get(service, url, headers=headers)
    assert resp.status == falcon.HTTP_UNAUTHORIZED


def test_user_id():
    uid = 101
    d = dict(uid=uid, groups=[])
    sid = sessionsdb.create(**d)

    headers = {"Authorization": sid}

    word = "hello"
    url = urls.echo + "/" + word
    assert hug.test.get(service, url, headers=headers).data == ("%s:%s" % (uid, word))

    url = "me/uid"

    data = {"uid": None}
    resp = hug.test.post(service, url, body=data, headers=headers)
    assert resp.data == uid

    data = {"uid": 1}  # invalid claim
    resp = hug.test.post(service, url, body=data, headers=headers)
    assert resp.data == uid

    data = {"uid": uid}
    resp = hug.test.post(service, url, body=data, headers=headers)
    assert resp.data == uid


def test_group_access():
    # 1. No group
    uid = 111
    groups = []
    d = dict(uid=uid, groups=groups)
    sid = sessionsdb.create(**d)
    url = urls.echo_for_groups

    headers = {"Authorization": sid}
    assert hug.test.get(service, url, headers=headers).status == falcon.HTTP_FORBIDDEN

    # 2. Forbidden group
    uid = 112
    groups = [globalgroups.forbidden.value]
    d = dict(uid=uid, groups=groups)
    sid = sessionsdb.create(**d)
    url = urls.echo_for_groups

    headers = {"Authorization": sid}
    assert hug.test.get(service, url, headers=headers).status == falcon.HTTP_FORBIDDEN

    # 3. Access group
    uid = 113
    groups = [globalgroups.privileged.value]
    d = dict(uid=uid, groups=groups)
    sid = sessionsdb.create(**d)

    headers = {"Authorization": sid}
    assert hug.test.get(service, url, headers=headers).status == falcon.HTTP_OK

    # 4. Other groups
    uid = 112
    groups = [globalgroups.others.value]
    d = dict(uid=uid, groups=groups)
    sid = sessionsdb.create(**d)
    url = urls.echo_for_groups

    headers = {"Authorization": sid}
    assert hug.test.get(service, url, headers=headers).status == falcon.HTTP_FORBIDDEN


def test_not_found():
    uid = 117
    d = dict(uid=uid, groups=[], site_groups={})
    sid = sessionsdb.create(**d)

    headers = {"Authorization": sid}

    url = "snakes/"
    assert hug.test.get(service, url).status == falcon.HTTP_NOT_FOUND

    url = "snakes/viper"
    resp = hug.test.get(service, url)
    assert resp.status == falcon.HTTP_OK
    assert resp.data == "viper"

    url = "sites/1/snakes/"
    assert hug.test.get(service, url, headers=headers).status == falcon.HTTP_NOT_FOUND

    url = "sites/1/snakes/viper"
    assert hug.test.get(service, url, headers=headers).status == falcon.HTTP_OK


def test_site_group_access():
    # 1. No group
    uid = 114
    groups = []
    site_groups = {}
    d = dict(uid=uid, groups=groups, site_groups=site_groups)
    sid = sessionsdb.create(**d)
    url = urls.echo_for_sitegroups

    headers = {"Authorization": sid}
    assert hug.test.get(service, url, headers=headers).status == falcon.HTTP_FORBIDDEN

    # 2. Forbidden group
    uid = 115
    groups = [globalgroups.forbidden.value]
    site_groups = {1: [sitegroups.privileged.value]}
    d = dict(uid=uid, groups=groups, site_groups=site_groups)
    sid = sessionsdb.create(**d)
    url = urls.echo_for_sitegroups

    headers = {"Authorization": sid}
    assert hug.test.get(service, url, headers=headers).status == falcon.HTTP_FORBIDDEN

    # 3. Access group
    uid = 116
    groups = [globalgroups.privileged.value]
    site_groups = {1: [sitegroups.privileged.value]}
    d = dict(uid=uid, groups=groups, site_groups=site_groups)
    sid = sessionsdb.create(**d)

    headers = {"Authorization": sid}
    assert hug.test.get(service, url, headers=headers).status == falcon.HTTP_OK


def test_all_site_group_access():
    url = urls.echo_for_all_groups

    # 1. No group
    uid = 114
    groups = []
    site_groups = {}
    d = dict(uid=uid, groups=groups, site_groups=site_groups)
    sid = sessionsdb.create(**d)

    headers = {"Authorization": sid}
    assert hug.test.get(service, url, headers=headers).status == falcon.HTTP_FORBIDDEN

    # 2. Forbidden group
    uid = 115
    groups = [globalgroups.forbidden.value]
    site_groups = {1: [sitegroups.forbidden.value]}
    d = dict(uid=uid, groups=groups, site_groups=site_groups)
    sid = sessionsdb.create(**d)

    headers = {"Authorization": sid}
    assert hug.test.get(service, url, headers=headers).status == falcon.HTTP_FORBIDDEN

    # 3. Access group
    uid = 116
    groups = [globalgroups.privileged.value]
    site_groups = {1: [sitegroups.privileged.value]}
    d = dict(uid=uid, groups=groups, site_groups=site_groups)
    sid = sessionsdb.create(**d)

    headers = {"Authorization": sid}
    assert hug.test.get(service, url, headers=headers).status == falcon.HTTP_OK


def test_bound_site_group_access():
    # 1. Forbidden group
    uid = 121
    groups = [globalgroups.forbidden.value]
    site_groups = {1: [sitegroups.forbidden.value]}
    d = dict(uid=uid, groups=groups, site_groups=site_groups, site_ctx=1)
    sid = sessionsdb.create(**d)
    url = urls.echo_for_sitegroups

    headers = {"Authorization": sid}
    assert hug.test.get(service, url, headers=headers).status == falcon.HTTP_FORBIDDEN

    # 2. Access group
    uid = 122
    groups = [globalgroups.privileged.value]
    site_groups = {1: [sitegroups.privileged.value]}
    d = dict(uid=uid, groups=groups, site_groups=site_groups, site_ctx=1)
    sid = sessionsdb.create(**d)

    headers = {"Authorization": sid}
    assert hug.test.get(service, url, headers=headers).status == falcon.HTTP_OK

    # 2. Access group of Unbound site
    uid = 123
    groups = [globalgroups.privileged.value]
    site_groups = {1: [sitegroups.privileged.value]}
    d = dict(uid=uid, groups=groups, site_groups=site_groups, site_ctx=2)
    sid = sessionsdb.create(**d)

    headers = {"Authorization": sid}
    assert (
        hug.test.get(service, url, headers=headers).status == falcon.HTTP_UNAUTHORIZED
    )

    uid = 123
    groups = [globalgroups.privileged.value]
    site_groups = {2: [sitegroups.privileged.value]}
    d = dict(uid=uid, groups=groups, site_groups=site_groups, site_ctx=1)
    sid = sessionsdb.create(**d)

    headers = {"Authorization": sid}
    assert hug.test.get(service, url, headers=headers).status == falcon.HTTP_FORBIDDEN


def test_request_access():
    url = "request-and-body"
    req = hug.test.post(
        service, url, body={"z": 1}, headers={"testheader": "testheader-value"}
    )
    resp = req.data
    assert "testheader".upper() in resp["headers"]
    assert resp["body"] == {"z": 1}


def test_raw_request():
    url = "request-raw-body"
    req = hug.test.post(
        service, url, body={"z": 1}, headers={"testheader": "testheader-value"}
    )
    resp = req.data
    assert "testheader".upper() in resp["headers"]


def test_custom_authorization_access():
    uid = 111
    groups = []
    d = dict(uid=uid, groups=groups)
    sid = sessionsdb.create(**d)
    headers = {"Authorization": sid}

    url = urls.echo_for_custom_authorization + "/authorized"
    assert hug.test.get(service, url).status == falcon.HTTP_UNAUTHORIZED

    url = urls.echo_for_custom_authorization + "/authorized"
    assert hug.test.get(service, url, headers=headers).status == falcon.HTTP_OK

    url = urls.echo_for_custom_authorization + "/unauthorized"
    assert hug.test.get(service, url, headers=headers).status == falcon.HTTP_FORBIDDEN


def test_honeybadger_wrapper():

    mocked_honeybadger = mock.MagicMock()
    wrapper = honeybadger_wrapper(mocked_honeybadger)

    def good_endpoint(foo):
        return foo

    class IgnorableError(BaseError):
        report = False

    def bad_endpoint(foo):
        raise IgnorableError()

    def worse_endpoint(foo, password):
        raise BaseError()

    def worst_endpoint(foo):
        raise RuntimeError()

    wrapped_good_endpoint = wrapper(good_endpoint)
    wrapped_bad_endpoint = wrapper(bad_endpoint)
    wrapped_worse_endpoint = wrapper(worse_endpoint)
    wrapped_worst_endpoint = wrapper(worst_endpoint)

    assert wrapped_good_endpoint(1) == 1
    assert not mocked_honeybadger.notify.called

    with pytest.raises(IgnorableError):
        wrapped_bad_endpoint(1)
    assert not mocked_honeybadger.notify.called

    with pytest.raises(BaseError) as e:
        wrapped_worse_endpoint(1, password="secret")
    mocked_honeybadger.notify.assert_called_once_with(
        e.value,
        context={
            "func": "worse_endpoint",
            "args": (1,),
            "kwargs": {"password": "[FILTERED]"},
        },
    )
    assert mocked_honeybadger.notify.call_count == 1

    with pytest.raises(RuntimeError):
        wrapped_worst_endpoint(1)
    assert mocked_honeybadger.notify.call_count == 2

    mocked_honeybadger.notify.side_effect = HTTPError(
        response=mock.MagicMock(status_code=403)
    )
    with pytest.raises(RuntimeError):
        wrapped_worst_endpoint(1)
    # TODO: How to check nested exception?
    assert mocked_honeybadger.notify.call_count == 3

    mocked_honeybadger.notify.side_effect = HTTPError(
        response=mock.MagicMock(status_code=401)
    )
    with pytest.raises(HTTPError):
        wrapped_worst_endpoint(1)
    # TODO: How to check nested exception?
    assert mocked_honeybadger.notify.call_count == 4



import os
import smtplib
import ssl
from email import encoders
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formataddr

import html2text
from converge import settings

from apphelpers.loggers import app_logger


def format_msg(
    sender,
    recipients,
    subject,
    text=None,
    html=None,
    attachments=None,
    images=None,
    reply_to=None,
    bcc=None,
    headers=None,
):
    if html and not text:
        text = html2text.html2text(html)

    from_header = sender
    if isinstance(sender, (list, tuple)):
        from_header = formataddr(sender)

    msg = MIMEMultipart("alternative")
    msg["Subject"] = subject
    msg["From"] = from_header
    msg["To"] = ", ".join(recipients)

    if bcc:
        msg["bcc"] = ", ".join(bcc)
    if reply_to:
        msg.add_header("reply-to", reply_to)

    msg.attach(MIMEText(text, "plain", "utf-8"))
    if html:
        msg.attach(MIMEText(html, "html", "utf-8"))
    if images:
        for cid, img in images:
            img_part = MIMEImage(img)
            img_part.add_header("Content-ID", "<" + cid + ">")
            msg.attach(img_part)
    if attachments:
        for file_path in attachments:
            file_name = os.path.basename(file_path)
            file_part = MIMEBase("application", "octet-stream")
            attachment = open(file_path, "rb")
            file_part.set_payload((attachment).read())
            encoders.encode_base64(file_part)
            file_part.add_header(
                "Content-Disposition", f"attachment; filename= {file_name}"
            )
            msg.attach(file_part)
    if headers:
        for key, value in headers.items():
            msg.add_header(key, value)
    return msg


def send_email(
    sender,
    recipients,
    subject,
    text=None,
    html=None,
    attachments=None,
    images=None,
    reply_to=None,
    bcc=None,
    headers=None,
):
    """
    text: text message. If html is provided and not text, text will be auto generated
    html: html message
    images: list of cid and image paths.
        eg. [('logo', 'images/logo.png'), ('Bruce', 'images/bat.png')]

    sender: can be sender email string e.g. 'foo@example.com' or
    list/tuple sender name and email  ('Foo', 'foo@example.com')
    headers: Dictionary of additional headers.
    """
    assert any((text, html)), "please provide html or text"

    if settings.DEBUG or settings.APP_MODE != "prod":
        # Make sure that we don't send emails to external emails in dev/stage
        filtered_recipients = []
        all_recipients = (tuple(recipients) + tuple(bcc)) if bcc else recipients
        for recpt in all_recipients:
            if isinstance(recpt, (list, tuple)):
                _, email = recpt
            else:
                email = recpt

            if any(
                email.endswith(f"@{domain}")
                for domain in settings.INTERNAL_EMAIL_DOMAINS
            ):
                filtered_recipients.append(recpt)
            else:
                internal_domains = ", ".join(settings.INTERNAL_EMAIL_DOMAINS)
                app_logger.info(
                    f"Skipping email to {email} as it does not end with any of"
                    f" {internal_domains}"
                )

        recipients = filtered_recipients

    msg = format_msg(
        sender,
        recipients,
        subject,
        text,
        html,
        attachments,
        images,
        reply_to,
        bcc,
        headers,
    )

    context = ssl.create_default_context()

    with smtplib.SMTP(settings.SMTP_HOST, settings.SMTP_PORT, timeout=3) as s:
        if settings.SMTP_USERNAME:
            s.ehlo()
            s.starttls(context=context)
            s.login(settings.SMTP_USERNAME, settings.SMTP_KEY)

        s.send_message(msg=msg)



# -*- coding: utf-8 -*-
from collections import namedtuple

import pytest

import apphelpers.sessions as sessionslib
import settings
from apphelpers.errors import InvalidSessionError

sessiondb_conn = dict(
    host=settings.SESSIONSDB_HOST,
    port=settings.SESSIONSDB_PORT,
    password=settings.SESSIONSDB_PASSWD,
    db=settings.SESSIONSDB_NO,
)
sessionsdb = sessionslib.SessionDBHandler(sessiondb_conn)
sessionsdb.destroy_all()


class data:
    pass


class state:
    pass


Session = namedtuple("Session", ["uid", "groups", "k", "v"])
data.session = Session(987651, ["admin", "member"], "foo", "bar")
data.session_email = "noone@example.com"
data.site_ctx = 123

# uid, groups, k, v = 987651, ['admin', 'member'], 'foo', 'bar'


def test_create():
    d = dict(
        uid=data.session.uid,
        groups=data.session.groups,
        extras=dict(email=data.session_email),
    )
    sid = sessionsdb.create(**d)
    assert len(sid) == 43
    sid_new = sessionsdb.create(data.session.uid, data.session.groups)
    assert sid == sid_new == sessionsdb.uid2sid(d["uid"])
    state.sid = sid
    assert sessionsdb.uid2bound_sids(d["uid"]) == []

    d = dict(
        uid=data.session.uid,
        groups=data.session.groups,
        extras=dict(email=data.session_email),
        site_ctx=data.site_ctx,
    )
    bound_sid = sessionsdb.create(**d)
    assert bound_sid != sid
    assert bound_sid == sessionsdb.uid2sid(d["uid"], data.site_ctx)
    assert sessionsdb.uid2bound_sids(d["uid"]) == [bound_sid]
    assert sessionsdb.exists(bound_sid)
    assert sessionsdb.uid2bound_site_ids(d["uid"]) == [data.site_ctx]

    sessionsdb.destroy_all_for_bound_site(data.site_ctx)
    assert not sessionsdb.exists(bound_sid)

    d = dict(
        uid=data.session.uid,
        groups=data.session.groups,
        extras=dict(email=data.session_email),
        site_ctx=data.site_ctx,
    )
    bound_sid = sessionsdb.create(**d)
    assert bound_sid != sid
    assert bound_sid == sessionsdb.uid2sid(d["uid"], data.site_ctx)
    assert sessionsdb.uid2bound_sids(d["uid"]) == [bound_sid]
    assert sessionsdb.exists(bound_sid)

    sessionsdb.destroy_bound_sessions_for(data.session.uid)
    assert not sessionsdb.exists(bound_sid)


def test_update():
    sid = state.sid
    k, v = data.session.k, data.session.v
    sessionsdb.update(sid, {k: v})
    d = sessionsdb.get(sid)
    assert d[k] == v
    assert d["email"] == data.session_email
    sessionsdb.remove_from_session(sid, [k])
    d = sessionsdb.get(sid)
    assert k not in d


def test_resync():
    sid = state.sid
    k, v = data.session.k, data.session.v
    sessionsdb.update(sid, {k: v})
    d = sessionsdb.get(sid)
    assert d[k] == v

    d = dict(
        uid=data.session.uid,
        groups=data.session.groups,
        extras=dict(email=data.session_email),
    )
    sessionsdb.resync(sid, d)
    d = sessionsdb.get(sid)
    assert k not in d


def test_delete():
    sessionsdb.destroy(state.sid)
    with pytest.raises(InvalidSessionError):
        sessionsdb.get(state.sid)


def test_session_lookups():
    uids = range(10000, 10010)
    groups = ["grp1", "grp2"]
    for uid in uids:
        sid = sessionsdb.create(uid, groups)
        assert sessionsdb.sid2uidgroups(sid) == (uid, groups)
        sessionsdb.destroy(sid)
        with pytest.raises(InvalidSessionError):
            sessionsdb.get(sid)



from dataclasses import dataclass
from typing import ClassVar

from fastapi import HTTPException, status


@dataclass
class BaseError(HTTPException):
    # Whether to report this error to honeybadger
    report: ClassVar[bool] = True
    status_code: ClassVar[int] = status.HTTP_500_INTERNAL_SERVER_ERROR
    detail: ClassVar[str] = "Something went wrong"


@dataclass
class HTTP400BadRequest(BaseError):
    status_code: ClassVar[int] = status.HTTP_400_BAD_REQUEST
    detail: str = "Bad Request"


@dataclass
class HTTP401Unauthorized(BaseError):
    status_code: ClassVar[int] = status.HTTP_401_UNAUTHORIZED
    detail: str = "Unauthorized"


@dataclass
class HTTP403Forbidden(BaseError):
    status_code: ClassVar[int] = status.HTTP_403_FORBIDDEN
    detail: str = "Forbidden"


@dataclass
class HTTP404NotFound(BaseError):
    status_code: ClassVar[int] = status.HTTP_404_NOT_FOUND
    detail: str = "Not Found"


@dataclass
class HTTP409Conflict(BaseError):
    status_code: ClassVar[int] = status.HTTP_409_CONFLICT
    detail: str = "Conflict"


@dataclass
class InvalidSessionError(HTTP401Unauthorized):
    detail: str = "Invalid Session"



import os
import sys

import fastapi

sys.path.append(os.getcwd() + "/..")

import settings  # noqa E402
from apphelpers.rest.fastapi import APIFactory  # noqa E402
from fastapi_tests.app.endpoints import setup_routes  # noqa E402
from fastapi_tests.app.models import db  # noqa E402


def make_app():
    sessiondb_conn = dict(
        host=settings.SESSIONSDB_HOST,
        port=settings.SESSIONSDB_PORT,
        password=settings.SESSIONSDB_PASSWD,
        db=settings.SESSIONSDB_NO,
    )

    api_factory = APIFactory(sessiondb_conn=sessiondb_conn, site_identifier="site_id")
    api_factory.setup_db_transaction(db)
    setup_routes(api_factory)

    app = fastapi.FastAPI()
    for router in api_factory.list_routers():
        app.include_router(router)
    api_factory.setup_session_db(sessiondb_conn)
    return app


app = make_app()



import os

from google.auth.transport import requests
from google.oauth2 import id_token
from requests_oauthlib import OAuth2Session

try:
    from converge import settings
except ImportError:
    import settings

if not settings.G_REDIRECT_URI.startswith("https"):
    os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "True"


def fetch_info(access_token):
    session = OAuth2Session(token={"access_token": access_token})
    userinfo = session.get("https://www.googleapis.com/oauth2/v1/userinfo").json()
    return userinfo


def fetch_info_using_jwt(access_token):
    request = requests.Request()
    userinfo = id_token.verify_oauth2_token(
        id_token=access_token, request=request, audience=settings.G_CLIENT_ID
    )
    return userinfo



from falcon import HTTPError, status_codes


class BaseError(HTTPError):
    # Whether to report this error to honeybadger
    report = True

    status = status_codes.HTTP_500
    description: str = "Something went wrong"

    def __init__(self, status=None, description=None):
        super().__init__(
            status=status or self.status,
            description=description or self.description,
        )

    def to_dict(self):
        return {
            "status": self.status,
            "description": self.description,
        }


class NotFoundError(BaseError):
    status = status_codes.HTTP_404
    description = "Not Found"


class AccessDenied(BaseError):
    status = status_codes.HTTP_403
    description = "Access denied"


class ValidationError(BaseError):
    status = status_codes.HTTP_400
    description = "Invalid request"


class InvalidSessionError(BaseError):
    status = status_codes.HTTP_401
    description = "Invalid session"


class ConflictError(BaseError):
    status = status_codes.HTTP_409
    description = "Duplicate resource"



"""
This file marks the root of tests.
Common pytest fixtures can be defines here.
See https://docs.pytest.org/en/latest/reference/fixtures.html
"""

import pytest
from httpx import ASGITransport, AsyncClient

import settings
from apphelpers.async_sessions import SessionDBHandler
from fastapi_tests import service


sessiondb_conn = dict(
    host=settings.SESSIONSDB_HOST,
    port=settings.SESSIONSDB_PORT,
    password=settings.SESSIONSDB_PASSWD,
    db=settings.SESSIONSDB_NO,
)


@pytest.fixture
async def sessionsdb():
    sessionsdb = SessionDBHandler(sessiondb_conn)
    yield sessionsdb
    await sessionsdb.close()


@pytest.fixture
async def client():
    async with AsyncClient(
        transport=ASGITransport(app=service.app), base_url="http://test"
    ) as ac:
        yield ac


@pytest.fixture(scope="session", autouse=True)
async def setup():
    from apphelpers.db.piccolo import destroy_db_from_basetable, setup_db_from_basetable
    from fastapi_tests.app.models import BaseTable

    sessionsdb = SessionDBHandler(sessiondb_conn)
    await sessionsdb.destroy_all()
    destroy_db_from_basetable(BaseTable)
    setup_db_from_basetable(BaseTable)



import os

from requests_oauthlib import OAuth2Session

try:
    from converge import settings
except ImportError:
    import settings

os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "True"


def fetch_info(access_token):
    session = OAuth2Session(token={"access_token": access_token})
    info_url = "https://graph.facebook.com/me?fields=" + settings.FB_USER_FIELDS
    return session.get(info_url).json()



import hug

import settings
from apphelpers.rest.hug import APIFactory
from tests.app.endpoints import setup_routes


def make_app():
    router = hug.route.API(__name__)

    api_factory = APIFactory(router)
    api_factory.enable_multi_site(site_identifier="site_id")
    # api_factory.setup_db_transaction(app.models.db)
    sessiondb_conn = dict(
        host=settings.SESSIONSDB_HOST,
        port=settings.SESSIONSDB_PORT,
        password=settings.SESSIONSDB_PASSWD,
        db=settings.SESSIONSDB_NO,
    )
    api_factory.setup_session_db(sessiondb_conn)
    setup_routes(api_factory)


make_app()



import hug
import hug.directives

from apphelpers.rest import endpoint as ep
from apphelpers.rest.hug import user_id
from tests.app.models import globalgroups, sitegroups


def echo(word, user: hug.directives.user = None):
    return "%s:%s" % (user.id, word) if user else word


@ep.login_required
def secure_echo(word, user: hug.directives.user = None):
    return "%s:%s" % (user.id, word) if user else word


@ep.any_group_required(globalgroups.privileged)
@ep.groups_forbidden(globalgroups.forbidden)
def echo_groups(user: hug.directives.user = None):
    return user.groups


def add(nums: hug.types.multiple):
    return sum(int(x) for x in nums)


@ep.login_required
def get_my_uid(uid: user_id):
    return uid


@ep.not_found_on_none
def get_snake(name=None):
    return name


@ep.login_required
@ep.not_found_on_none
def get_secure_snake(site_id, name=None):
    return name


@ep.login_required
def secure_multisite_echo(word, user: hug.directives.user = None):
    return "%s:%s" % (user.id, word) if user else word


@ep.any_group_required(sitegroups.privileged)
@ep.groups_forbidden(globalgroups.forbidden)
def echo_multisite_groups(site_id: int, user: hug.directives.user = None):
    return user.groups


@ep.all_groups_required({globalgroups.privileged, sitegroups.privileged})
def echo_multisite_all_groups(site_id: int, user: hug.directives.user = None):
    return user.groups + user.site_groups[site_id]


def process_request(request, body):
    return {"body": body, "headers": request.headers}


@ep.not_found_on_none
def process_raw_request(request):
    return {"raw_body": request.stream.read().decode(), "headers": request.headers}


def check_authorization(user, *args, **kw):
    return kw["word"] == "authorized"


@ep.authorizer(check_authorization)
def custom_authorization_echo(word):
    return word


def setup_routes(factory):
    factory.get("/echo/{word}")(echo)
    factory.post("/echo")(echo)

    factory.get("/add")(add)

    factory.get("/secure-echo/{word}")(secure_echo)
    factory.get("/custom-authorization-echo/{word}")(custom_authorization_echo)
    factory.get("/echo-groups")(echo_groups)

    factory.post("/me/uid")(get_my_uid)

    factory.get("/snakes/{name}")(get_snake)

    factory.get("/sites/{site_id}/secure-echo/{word}")(secure_multisite_echo)
    factory.get("/sites/{site_id}/echo-groups")(echo_multisite_groups)
    factory.get("/sites/{site_id}/echo-all-groups")(echo_multisite_all_groups)
    factory.get("/sites/{site_id}/snakes/{name}")(get_secure_snake)

    factory.post("/request-and-body")(process_request)
    factory.post("/request-raw-body", parse_body=False)(process_raw_request)

    # ar_handlers = (None, arlib.create, None, arlib.get, arlib.update, None)
    # factory.map_resource('/resttest/', handlers=ar_handlers)



import datetime
import logging
import os
from contextlib import contextmanager
from enum import Enum

import pytz
from peewee import Field, Model
from playhouse.pool import PooledPostgresqlExtDatabase
from playhouse.postgres_ext import DateTimeTZField
from playhouse.shortcuts import model_to_dict

try:
    from hug.decorators import wraps
except ModuleNotFoundError:
    from functools import wraps


def set_peewee_debug():
    logger = logging.getLogger("peewee")
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(logging.DEBUG)


class URLField(Field):
    field_type = "text"

    def python_value(self, value):
        value = value.lower()
        if not value.startswith("http") or "://" not in value:
            raise TypeError("Not a http URL: %s" % value)
        return value


def create_pgdb_pool(
    host=None,
    database=None,
    user=None,
    password=None,
    port=None,
    sslmode=None,
    max_connections=32,
):
    return PooledPostgresqlExtDatabase(
        database=database,
        host=host,
        user=user,
        password=password,
        port=port,
        sslmode=sslmode,
        max_connections=max_connections,
        register_hstore=False,
        stale_timeout=60 * 2,
    )  # 2 minutes


def create_base_model(db):
    class BaseModel(Model):
        class Meta:
            database = db
            only_save_dirty = True

        def to_dict(self, only=None, exclude=None, recurse=False):
            return model_to_dict(
                self,
                only=only,
                exclude=exclude,
                recurse=recurse,
            )

    return BaseModel


def created():
    return DateTimeTZField(default=lambda: datetime.datetime.now(pytz.utc))


@contextmanager
def dbtransaction_ctx(db):
    if not db.in_transaction():
        with db.connection_context():
            with db.atomic():
                yield
    else:
        with db.atomic():
            yield


def dbtransaction(db):
    """
    wrapper that make db transactions automic
    note db connections are used only when it is needed (hence there is no
    usual connection open/close)
    """

    def wrapper(f):
        @wraps(f)
        def f_wrapped(*args, **kw):
            with dbtransaction_ctx(db):
                return f(*args, **kw)

        return f_wrapped

    return wrapper


def enumify(TheModel, name_field="name", val_field="id"):
    """
    Converts a model rows into an enum
    Can be effective cache for mostly unchanging data.
    Limitation: No auto updates. If you update the model and you are using
    process manager like gunicorn you would need to restart to rnsure enums
    are updated

    eg.
    >>> class Week(BaseModel):
            day = CharField()
            num = IntField()

    >>> weekenum = enumify(Week, 'day', 'num')
    >>> print(weekenum.monday.num)
    """
    fields = getattr(TheModel, name_field), getattr(TheModel, val_field)
    data = list(
        (name.replace(" ", "_").lower(), v)
        for (name, v) in TheModel.select(*fields).tuples()
    )
    return Enum(TheModel.__name__, data)


def dbc(db):
    pid = os.getpid()
    return "[%s]: %s:%s" % (pid, len(db._in_use), db.max_connections)


def get_sub_models(base_model):
    models = []
    for sub_model in base_model.__subclasses__():
        models.append(sub_model)
        models.extend(get_sub_models(sub_model))
    return models


# Useful functions for test/dev setups
# NOTE: For below functions, models is list of model classes sorted by
# dependency
# Example: [Author, Publication, Post, Comment]


def setup_db(db, models):
    db.create_tables(models, safe=True)


def setup_db_from_basemodel(db, basemodel):
    models = get_sub_models(basemodel)
    db.create_tables(models, safe=True)


def destroy_db(models):
    for o in models[::-1]:
        if o.table_exists():
            o.drop_table()
            print("DROP: " + o._meta.name)



from __future__ import annotations

import secrets
from typing import Any

import _pickle as pickle
import redis

from apphelpers.errors import InvalidSessionError

_SEP = ":"
session_key = ("session" + _SEP).__add__

rev_lookup_prefix = f"uid{_SEP}"


def rev_lookup_key(uid, site_ctx=None):
    return (
        f"{rev_lookup_prefix}{uid}{_SEP}{site_ctx}"
        if site_ctx
        else f"{rev_lookup_prefix}{uid}"
    )


THIRTY_DAYS = 30 * 24 * 60 * 60


class SessionDBHandler:
    def __init__(self, rconn_params):
        """
        rconn_params: redis connection parameters
        """
        self.rconn = redis.Redis(**rconn_params)

    def create(
        self,
        uid="",
        groups=None,
        site_groups=None,
        extras=None,
        ttl=THIRTY_DAYS,
        site_ctx=None,
    ):
        """
        groups: list
        site_groups: dict
        extras (dict): each key-value pair of extras get stored into hset
        site_ctx: int (session is only applicable for bound site_id)
        """
        if uid:
            sid = self.uid2sid(uid, site_ctx)
            if sid:
                return sid

        sid = secrets.token_urlsafe()
        key = session_key(sid)

        if groups is None:
            groups = []
        session_dict = {
            "uid": uid,
            "groups": groups,
            "site_groups": site_groups,
            "site_ctx": site_ctx,
        }
        if extras:
            session_dict.update(extras)
        session = {k: pickle.dumps(v) for k, v in session_dict.items()}
        self.rconn.hset(key, mapping=session)

        if uid:
            rev_key = rev_lookup_key(uid, site_ctx)
            self.rconn.setex(rev_key, value=sid, time=ttl)
        self.rconn.expire(key, ttl)
        return sid

    def exists(self, sid):
        return self.rconn.exists(session_key(sid))

    def get(self, sid, keys=[]) -> dict[str, Any]:
        s_values = self.rconn.hgetall(session_key(sid))
        if not s_values:
            raise InvalidSessionError()
        session = {k.decode(): pickle.loads(v) for k, v in s_values.items()}
        if keys:
            session = {k: session.get(k, None) for k in keys}
        return session

    def get_attribute(self, sid, attribute):
        value = self.rconn.hget(session_key(sid), attribute)
        return pickle.loads(value) if value else None

    def uid2sid(self, uid, site_ctx=None):
        sid = self.rconn.get(rev_lookup_key(uid, site_ctx))
        return sid.decode() if sid else None

    def uid2bound_sids(self, uid):
        keys = self.rconn.keys(rev_lookup_key(uid, "*"))
        return [self.rconn.get(key).decode() for key in keys]

    def uid2bound_site_ids(self, uid):
        keys = self.rconn.keys(rev_lookup_key(uid, "*"))
        return [int(key.decode().split(_SEP)[2]) for key in keys]

    def sid2uid(self, sid):
        session = self.get(sid, ["uid"])
        return session["uid"]

    def get_for(self, uid):
        sid = self.uid2sid(uid)
        return self.get(sid) if sid else None

    def get_bound_sessions_for(self, uid):
        return [self.get(sid) for sid in self.uid2bound_sids(uid)]

    # Same default ttl as `create` function
    def extend_timeout(self, sid, ttl=THIRTY_DAYS):
        self.rconn.expire(session_key(sid), ttl)
        uid = self.sid2uid(sid)
        self.rconn.expire(rev_lookup_key(uid), ttl)

    def sid2uidgroups(self, sid):
        """
        => uid (int), groups (list)
        """
        session = self.get(sid, ["uid", "groups"])
        return session["uid"], session["groups"]

    def update(self, sid, keyvalues):
        sk = session_key(sid)
        keyvalues = {k: pickle.dumps(v) for k, v in list(keyvalues.items())}
        self.rconn.hset(sk, mapping=keyvalues)

    def update_for(self, uid, keyvalues):
        sid = self.uid2sid(uid)
        return self.update(sid, keyvalues) if sid else None

    def update_attribute(self, sid, attribute, value):
        key = session_key(sid)
        self.rconn.hset(key, attribute, pickle.dumps(value))
        return True

    def resync(self, sid, keyvalues):
        removed_keys = list(self.get(sid).keys() - keyvalues.keys())
        self.remove_from_session(sid, removed_keys)
        self.update(sid, keyvalues)

    def resync_for(self, uid, keyvalues, site_ctx=None):
        keyvalues["uid"] = uid
        keyvalues["site_ctx"] = site_ctx
        sid = self.uid2sid(uid, site_ctx)
        return self.resync(sid, keyvalues) if sid else None

    def remove_from_session(self, sid, keys):
        sk = session_key(sid)
        if keys:
            self.rconn.hdel(sk, *keys)
        return True

    def destroy(self, sid, site_ctx=None):
        uid = self.sid2uidgroups(sid)[0]
        sk = session_key(sid)
        self.rconn.delete(sk)
        self.rconn.delete(rev_lookup_key(uid, site_ctx))
        return True

    def destroy_for(self, uid, site_ctx=None):
        sid = self.uid2sid(uid, site_ctx)
        return self.destroy(sid, site_ctx) if sid else None

    def destroy_all(self):
        keys = self.rconn.keys(session_key("*"))
        if keys:
            self.rconn.delete(*keys)
        keys = self.rconn.keys(rev_lookup_prefix + "*")
        if keys:
            self.rconn.delete(*keys)

    def destroy_all_for_bound_site(self, site_ctx):
        keys = self.rconn.keys(rev_lookup_key("*", site_ctx))
        if keys:
            sids = [session_key(self.rconn.get(key).decode()) for key in keys]
            if sids:
                self.rconn.delete(*sids)
            self.rconn.delete(*keys)

    def destroy_bound_sessions_for(self, uid):
        keys = self.rconn.keys(rev_lookup_key(uid, "*"))
        if keys:
            sids = [session_key(self.rconn.get(key).decode()) for key in keys]
            if sids:
                self.rconn.delete(*sids)
            self.rconn.delete(*keys)



import logging
import os
from logging.handlers import RotatingFileHandler

from converge import settings
from loguru import logger as loguru_logger


def build_api_logger(rotate=False):
    """
    Builds multiprocess-safe (hence loguru over stdlib logger) API logger
    """
    handler = settings.API_LOGGER.FILEPATH
    if handler:  # Else log to sys.stderr by default
        loguru_logger.add(
            handler,
            retention=settings.API_LOGGER.RETENTION if rotate else None,
            rotation=settings.API_LOGGER.ROTATION if rotate else None,
            format="{time:YYYY-MM-DD HH:mm:ss} | {message}",
            enqueue=True,
            level=settings.API_LOGGER.LEVEL,
        )
    return loguru_logger


def build_app_logger(name="app", logfile="app.log", debug=True, rotate=False):
    """
    General purpose application logger. Useful mainly for debugging
    """
    level = logging.DEBUG if settings.DEBUG else logging.INFO
    logdir = getattr(getattr(settings, "APP_LOGGER", None), "LOGDIR", "logs")
    if not os.path.exists(logdir):
        os.mkdir(logdir)
    logpath = os.path.join(logdir, logfile)
    maxBytes = 1024 * 1024 * 10

    if rotate:
        handler = RotatingFileHandler(logpath, maxBytes=maxBytes, backupCount=100)
    else:
        handler = logging.FileHandler(logpath)
    handler.setLevel(level)
    formatter = logging.Formatter("[%(levelname)s] %(asctime)s: %(message)s")
    handler.setFormatter(formatter)
    logger = logging.getLogger(name)
    logger.addHandler(handler)
    logger.setLevel(level)
    return logger


app_logger = build_app_logger()
app_logger.info(f"Running app in '{settings.APP_MODE}' mode")

api_logger = None
if settings.API_LOGGER.ENABLED:
    api_logger = build_api_logger()
    api_logger.info(f"Running api in '{settings.APP_MODE}' mode")



try:
    from apphelpers.errors.fastapi import *  # noqa: F401, F403

    print("apphelpers: using FastAPI")
except ImportError:
    from apphelpers.errors.hug import *  # noqa: F401, F403

    print("apphelpers: using hug")



import asyncio

from piccolo import columns as col
from piccolo.engine.postgres import PostgresEngine
import pytest

import settings
from apphelpers.db.piccolo import (
    BaseTable,
    dbtransaction,
    destroy_db_from_basetable,
    setup_db_from_basetable,
)

db = PostgresEngine(
    config=dict(
        host=settings.DB_HOST,
        database=settings.DB_NAME,
        user=settings.DB_USER,
        password=settings.DB_PASS,
    )
)
dbtransaction_wrapper = dbtransaction(db)


class Book(BaseTable, db=db):
    name = col.Text()


async def _add_book(name):
    await Book.insert(Book(name=name)).run()


async def _add_book_loser(name):
    await _add_book(name)
    loser  # will raise  # noqa: F821


def setup_module():
    setup_db_from_basetable(BaseTable)


def teardown_module():
    destroy_db_from_basetable(BaseTable)


async def add_with_tr():
    add_book = dbtransaction_wrapper(_add_book)
    name = "The Pillars of the Earth"
    await add_book(name)
    names = [b[Book.name] for b in await Book.select().run()]
    assert name in names

    add_book_loser = dbtransaction_wrapper(_add_book_loser)
    name = "The Cathedral and the Bazaar"
    with pytest.raises(NameError):
        await add_book_loser(name)
    names = [b[Book.name] for b in await Book.select().run()]
    assert name not in names

    add_book = dbtransaction_wrapper(_add_book)
    name = "The Ego Trick"
    await add_book(name)
    names = [b[Book.name] for b in await Book.select().run()]
    assert name in names


def test_add_with_tr():
    asyncio.run(add_with_tr())



import inspect
from functools import wraps
from typing import Annotated

from converge import settings
from fastapi import APIRouter, Depends, Header, Body, File, Query, Form
from fastapi.routing import APIRoute
from starlette.requests import Request

from apphelpers.db import dbtransaction_ctx, peewee_enabled
from apphelpers.errors.fastapi import (
    BaseError,
    HTTP401Unauthorized,
    HTTP403Forbidden,
    HTTP404NotFound,
    InvalidSessionError,
)
from apphelpers.rest import endpoint as ep
from apphelpers.rest.common import User, notify_honeybadger, phony
from apphelpers.async_sessions import SessionDBHandler

if settings.get("HONEYBADGER_API_KEY"):
    from honeybadger import Honeybadger


def raise_not_found_on_none(f):
    if getattr(f, "not_found_on_none", None) is True:
        if inspect.iscoroutinefunction(f):

            @wraps(f)
            async def async_wrapper(*ar, **kw):
                ret = await f(*ar, **kw)
                if ret is None:
                    raise HTTP404NotFound()
                return ret

            return async_wrapper
        else:

            @wraps(f)
            def wrapper(*ar, **kw):
                ret = f(*ar, **kw)
                if ret is None:
                    raise HTTP404NotFound()
                return ret

            return wrapper
    return f


def honeybadger_wrapper(hb):
    """
    wrapper that executes the function in a try/except
    If an exception occurs, it is first reported to Honeybadger
    """

    def wrapper(f):
        if inspect.iscoroutinefunction(f):

            @wraps(f)
            async def async_f_wrapped(*args, **kw):
                err_to_report = None
                try:
                    return await f(*args, **kw)
                except BaseError as e:
                    if e.report:
                        err_to_report = e
                    raise e
                except Exception as e:
                    err_to_report = e
                    raise e
                finally:
                    if err_to_report:
                        notify_honeybadger(
                            honeybadger=hb,
                            error=err_to_report,
                            func=f,
                            args=args,
                            kwargs=kw,
                        )

            return async_f_wrapped

        else:

            @wraps(f)
            def f_wrapped(*args, **kw):
                err_to_report = None
                try:
                    return f(*args, **kw)
                except BaseError as e:
                    if e.report:
                        err_to_report = e
                    raise e
                except Exception as e:
                    err_to_report = e
                    raise e

                finally:
                    if err_to_report:
                        notify_honeybadger(
                            honeybadger=hb,
                            error=err_to_report,
                            func=f,
                            args=args,
                            kwargs=kw,
                        )

            return f_wrapped

    return wrapper


if peewee_enabled:

    def dbtransaction(db):
        """
        wrapper that make db transactions automic
        note db connections are used only when it is needed (hence there is no
        usual connection open/close)
        """

        def wrapper(f):
            if inspect.iscoroutinefunction(f):

                @wraps(f)
                async def async_wrapper(*ar, **kw):
                    with dbtransaction_ctx(db):
                        return await f(*ar, **kw)

                return async_wrapper
            else:

                @wraps(f)
                async def sync_wrapper(*ar, **kw):
                    with dbtransaction_ctx(db):
                        return f(*ar, **kw)

                return sync_wrapper

        return wrapper

else:
    # for piccolo db
    def dbtransaction(engine, allow_nested=True):
        async def dependency():
            async with dbtransaction_ctx(engine, allow_nested=allow_nested):
                yield

        return Depends(dependency)


async def get_current_user(request: Request):
    return request.state.user if request.state.user.id else None


async def get_current_user_id(request: Request):
    return request.state.user.id


async def get_current_user_name(request: Request):
    return request.state.user.name


async def get_current_user_email(request: Request):
    return request.state.user.email


async def get_current_user_mobile(request: Request):
    return request.state.user.mobile


async def get_current_domain(request: Request):
    return request.headers["HOST"]


async def get_json_body(request: Request):
    return (
        await request.json()
        if request.headers.get("content-type") == "application/json"
        else {}
    )


async def get_raw_body(request: Request):
    return await request.body()


async def get_user_agent(request: Request):
    return request.headers.get("USER-AGENT", "")


async def get_user_ip(request: Request):
    return request.headers["X-FORWARDED-FOR"]


raw_body = Annotated[bytes, Depends(get_raw_body)]
json_body = Annotated[dict, Depends(get_json_body)]


class RequestHeaders:
    user_agent = Annotated[str, Depends(get_user_agent)]
    user_ip = Annotated[str, Depends(get_user_ip)]
    domain = Annotated[str, Depends(get_current_domain)]
    custom = Annotated[str, Header()]


class AuthParams:
    user = Annotated[User, Depends(get_current_user)]
    user_id = Annotated[int, Depends(get_current_user_id)]
    user_name = Annotated[str, Depends(get_current_user_name)]
    user_email = Annotated[str, Depends(get_current_user_email)]
    user_mobile = Annotated[str, Depends(get_current_user_mobile)]


class QueryParams:
    list_of_str = Annotated[list[str], Query()]
    list_of_int = Annotated[list[int], Query()]


class BodyParams:
    str = Annotated[str, Body(embed=True)]
    int = Annotated[int, Body(embed=True)]
    file = Annotated[bytes, File()]
    float = Annotated[float, Body(embed=True)]
    bool = Annotated[bool, Body(embed=True)]
    dict = Annotated[dict, Body(embed=True)]
    bytes = Annotated[bytes, Body(embed=True)]
    list_of_str = Annotated[list[str], Body(embed=True)]
    list_of_int = Annotated[list[int], Body(embed=True)]
    list_of_float = Annotated[list[float], Body(embed=True)]
    list_of_bool = Annotated[list[bool], Body(embed=True)]
    list_of_dict = Annotated[list[dict], Body(embed=True)]


class FormParams:
    str = Annotated[str, Form()]
    int = Annotated[int, Form()]
    file = Annotated[bytes, File()]
    float = Annotated[float, Form()]
    bool = Annotated[bool, Form()]
    bytes = Annotated[bytes, Form()]


class AuthByHeaderRouter(APIRoute):
    sessions = None
    auth_header_name = None

    @classmethod
    def setup_sessions(cls, sessions: SessionDBHandler):
        cls.sessions = sessions

    @classmethod
    def set_auth_header_name(cls, name: str):
        cls.auth_header_name = name

    def extract_auth_token(self, _request: Request):
        return _request.headers.get(self.auth_header_name)

    def get_route_handler(self):
        original_route_handler = super().get_route_handler()

        async def custom_route_handler(_request: Request):
            uid, groups, name, email, mobile, site_groups, site_ctx = (
                None,
                [],
                "",
                None,
                None,
                {},
                None,
            )

            token = self.extract_auth_token(_request)
            if token:
                session = await self.sessions.get(  # type: ignore
                    token,
                    [
                        "uid",
                        "name",
                        "groups",
                        "email",
                        "mobile",
                        "site_groups",
                        "site_ctx",
                    ],
                )
                uid, name, groups, email, mobile, site_groups, site_ctx = (
                    session["uid"],
                    session["name"],
                    session["groups"],
                    session["email"],
                    session["mobile"],
                    session["site_groups"],
                    session["site_ctx"],
                )

            _request.state.user = User(
                sid=token,
                id=uid,
                name=name,
                groups=groups,
                email=email,
                mobile=mobile,
                site_groups=site_groups,
                site_ctx=site_ctx,
            )

            return await original_route_handler(_request)

        original_route_handler.__signature__ = inspect.signature(
            original_route_handler
        ).replace(
            parameters=[
                # Use all parameters from handler
                *inspect.signature(original_route_handler).parameters.values(),
                inspect.Parameter(
                    name="_request",
                    kind=inspect.Parameter.VAR_POSITIONAL,
                    annotation=Request,
                ),
            ],
        )
        return custom_route_handler


class AuthByCookieOrHeaderRouter(AuthByHeaderRouter):
    auth_cookie_name = None

    @classmethod
    def set_auth_cookie_name(cls, cookie_name: str):
        cls.auth_cookie_name = cookie_name

    def extract_auth_token(self, _request: Request):
        return _request.cookies.get(self.auth_cookie_name) or _request.headers.get(
            self.auth_header_name
        )


class OptionalAuthByHeaderRouter(AuthByHeaderRouter):
    def get_route_handler(self):
        original_route_handler = super(AuthByHeaderRouter, self).get_route_handler()

        async def custom_route_handler(_request: Request):
            uid, groups, name, email, mobile, site_groups, site_ctx = (
                None,
                [],
                "",
                None,
                None,
                {},
                None,
            )

            token = self.extract_auth_token(_request)
            if token:
                try:
                    session = await self.sessions.get(  # type: ignore
                        token,
                        [
                            "uid",
                            "name",
                            "groups",
                            "email",
                            "mobile",
                            "site_groups",
                            "site_ctx",
                        ],
                    )
                    uid, name, groups, email, mobile, site_groups, site_ctx = (
                        session["uid"],
                        session["name"],
                        session["groups"],
                        session["email"],
                        session["mobile"],
                        session["site_groups"],
                        session["site_ctx"],
                    )
                except InvalidSessionError:
                    pass

            _request.state.user = User(
                sid=token,
                id=uid,
                name=name,
                groups=groups,
                email=email,
                mobile=mobile,
                site_groups=site_groups,
                site_ctx=site_ctx,
            )
            return await original_route_handler(_request)

        original_route_handler.__signature__ = inspect.signature(
            original_route_handler
        ).replace(
            parameters=[
                # Use all parameters from handler
                *inspect.signature(original_route_handler).parameters.values(),
                inspect.Parameter(
                    name="_request",
                    kind=inspect.Parameter.VAR_POSITIONAL,
                    annotation=Request,
                ),
            ],
        )
        return custom_route_handler


class OptionalAuthByCookieOrHeaderRouter(OptionalAuthByHeaderRouter):
    auth_cookie_name = None

    @classmethod
    def set_auth_cookie_name(cls, cookie_name: str):
        cls.auth_cookie_name = cookie_name

    def extract_auth_token(self, _request: Request):
        return _request.cookies.get(self.auth_cookie_name) or _request.headers.get(
            self.auth_header_name
        )


class APIFactory:
    def __init__(
        self,
        sessiondb_conn=None,
        urls_prefix="",
        site_identifier=None,
        auth_header_name="Authorization",
        auth_cookie_name="__s",
    ):
        self.access_wrapper = phony
        self.multi_site_enabled = False
        self.site_identifier = site_identifier
        self.urls_prefix = urls_prefix.rstrip("/")
        self.db_tr_wrapper = phony
        self.honeybadger_wrapper = phony

        self.sessions = None
        self.sessiondb_conn = sessiondb_conn
        if site_identifier:
            self.enable_multi_site(site_identifier)
        if sessiondb_conn:
            self.setup_session_db(sessiondb_conn)

        self.unsecure_router = APIRouter()
        self.auth_by_header_router = APIRouter(route_class=AuthByHeaderRouter)
        self.auth_by_cookie_or_header_router = APIRouter(
            route_class=AuthByCookieOrHeaderRouter
        )
        self.optional_auth_by_header_router = APIRouter(
            route_class=OptionalAuthByHeaderRouter
        )
        self.optional_auth_by_cookie_or_header_router = APIRouter(
            route_class=OptionalAuthByCookieOrHeaderRouter
        )

        # For piccolo, dbtransaction is handled as dependency, we need separate
        # routers with dbtransaction dependency if setup_db_transaction is called
        self.unsecure_router_with_dbtransaction = None
        self.auth_by_header_router_with_dbtransaction = None
        self.auth_by_cookie_or_header_router_with_dbtransaction = None
        self.optional_auth_by_header_router_with_dbtransaction = None
        self.optional_auth_by_cookie_or_header_router_with_dbtransaction = None

        self.setup_auth_header(auth_header_name)
        self.setup_auth_cookie(auth_cookie_name)

    def list_routers(self) -> list[APIRouter]:
        # NOTE Here order of router inclusion is important.
        # The secure router should be included before the router.
        return [
            router
            for router in (
                self.auth_by_header_router,
                self.auth_by_header_router_with_dbtransaction,
                self.auth_by_cookie_or_header_router,
                self.auth_by_cookie_or_header_router_with_dbtransaction,
                self.optional_auth_by_header_router,
                self.optional_auth_by_header_router_with_dbtransaction,
                self.optional_auth_by_cookie_or_header_router,
                self.optional_auth_by_cookie_or_header_router_with_dbtransaction,
                self.unsecure_router,
                self.unsecure_router_with_dbtransaction,
            )
            if router is not None
        ]

    def enable_multi_site(self, site_identifier: str):
        self.multi_site_enabled = True
        self.site_identifier = site_identifier
        if self.sessions:
            # if sessions are already set up, we need to reconfigure them
            self.setup_session_db(self.sessiondb_conn)

    def setup_db_transaction(self, db):
        if peewee_enabled:
            self.db_tr_wrapper = dbtransaction(db)
        else:
            # For piccolo, dbtransaction is handled as dependency,
            # we need separate routers with dbtransaction dependency
            self.unsecure_router_with_dbtransaction = APIRouter(
                dependencies=[dbtransaction(db)]
            )
            self.auth_by_header_router_with_dbtransaction = APIRouter(
                route_class=AuthByHeaderRouter, dependencies=[dbtransaction(db)]
            )
            self.auth_by_cookie_or_header_router_with_dbtransaction = APIRouter(
                route_class=AuthByCookieOrHeaderRouter,
                dependencies=[dbtransaction(db)],
            )
            self.optional_auth_by_header_router_with_dbtransaction = APIRouter(
                route_class=OptionalAuthByHeaderRouter, dependencies=[dbtransaction(db)]
            )
            self.optional_auth_by_cookie_or_header_router_with_dbtransaction = (
                APIRouter(
                    route_class=OptionalAuthByCookieOrHeaderRouter,
                    dependencies=[dbtransaction(db)],
                )
            )

    def setup_honeybadger_monitoring(self):
        api_key = settings.HONEYBADGER_API_KEY
        if not api_key:
            print("Info: Honeybadger API KEY not found. Honeybadger not set")
            return

        print("Info: Setting up Honeybadger")
        hb = Honeybadger()
        hb.configure(api_key=api_key)
        self.honeybadger_wrapper = honeybadger_wrapper(hb)

    def setup_auth_header(self, auth_header_name: str):
        AuthByHeaderRouter.set_auth_header_name(auth_header_name)
        AuthByCookieOrHeaderRouter.set_auth_header_name(auth_header_name)
        OptionalAuthByHeaderRouter.set_auth_header_name(auth_header_name)
        OptionalAuthByCookieOrHeaderRouter.set_auth_header_name(auth_header_name)

    def setup_auth_cookie(self, auth_cookie_name: str):
        AuthByCookieOrHeaderRouter.set_auth_cookie_name(auth_cookie_name)
        OptionalAuthByCookieOrHeaderRouter.set_auth_cookie_name(auth_cookie_name)

    def setup_session_db(self, sessiondb_conn):
        """
        redis_conn_params: dict() with below keys
                           (host, port, password, db)
        """
        self.sessions = SessionDBHandler(sessiondb_conn)
        OptionalAuthByHeaderRouter.setup_sessions(self.sessions)
        OptionalAuthByCookieOrHeaderRouter.setup_sessions(self.sessions)
        AuthByHeaderRouter.setup_sessions(self.sessions)
        AuthByCookieOrHeaderRouter.setup_sessions(self.sessions)

        def access_wrapper(f):
            """
            This is the authentication + authorization part
            """
            login_required = getattr(f, "login_required", None)
            any_group_required = getattr(f, "any_group_required", None)
            all_groups_required = getattr(f, "all_groups_required", None)
            groups_forbidden = getattr(f, "groups_forbidden", None)
            authorizer = getattr(f, "authorizer", None)

            if login_required:

                @wraps(f)
                async def wrapper(_request, *args, **kw):
                    user = _request.state.user

                    # this is authentication part
                    if not user.id:
                        raise HTTP401Unauthorized("Invalid or expired session")

                    # this is authorization part
                    groups = set(user.groups)

                    if any_group_required and groups.isdisjoint(any_group_required):
                        raise HTTP403Forbidden("Unauthorized access")

                    if all_groups_required and not groups.issuperset(
                        all_groups_required
                    ):
                        raise HTTP403Forbidden("Unauthorized access")

                    if groups_forbidden and groups.intersection(groups_forbidden):
                        raise HTTP403Forbidden("Unauthorized access")

                    if authorizer and not authorizer(user, *args, **kw):
                        raise HTTP403Forbidden("Unauthorized access")

                    return (
                        await f(*args, **kw)
                        if inspect.iscoroutinefunction(f)
                        else f(*args, **kw)
                    )

                f.__signature__ = inspect.signature(f).replace(
                    parameters=[
                        # Use all parameters from handler
                        *inspect.signature(f).parameters.values(),
                        inspect.Parameter(
                            name="_request",
                            kind=inspect.Parameter.VAR_POSITIONAL,
                            annotation=Request,
                        ),
                    ],
                )
            else:
                wrapper = f

            return wrapper

        def multisite_access_wrapper(f):
            """
            This is the authentication + authorization part
            """
            login_required = getattr(f, "login_required", False)
            any_group_required = getattr(f, "any_group_required", False)
            all_groups_required = getattr(f, "all_groups_required", False)
            groups_forbidden = getattr(f, "groups_forbidden", False)
            authorizer = getattr(f, "authorizer", False)

            if login_required:

                @wraps(f)
                async def wrapper(_request, *args, **kw):
                    user: User = _request.state.user
                    site_id = (
                        int(kw[self.site_identifier])
                        if kw.get(self.site_identifier) is not None
                        else None
                    )

                    # this is authentication part
                    if not user.id:
                        raise HTTP401Unauthorized("Invalid or expired session")

                    # bound site authorization
                    if (
                        user.site_ctx
                        and site_id != user.site_ctx
                        and getattr(f, "ignore_site_ctx", False) is False
                    ):
                        raise HTTP401Unauthorized("Invalid or expired session")

                    # this is authorization part
                    groups = set(user.groups)
                    if site_id:
                        groups = groups.union(user.site_groups.get(site_id, []))

                    if any_group_required and groups.isdisjoint(any_group_required):
                        raise HTTP403Forbidden("Unauthorized access")

                    if all_groups_required and not groups.issuperset(
                        all_groups_required
                    ):
                        raise HTTP403Forbidden("Unauthorized access")

                    if groups_forbidden and groups.intersection(groups_forbidden):
                        raise HTTP403Forbidden("Unauthorized access")

                    if authorizer and not authorizer(user, *args, **kw):
                        raise HTTP403Forbidden("Unauthorized access")

                    return (
                        await f(*args, **kw)
                        if inspect.iscoroutinefunction(f)
                        else f(*args, **kw)
                    )

                f.__signature__ = inspect.signature(f).replace(
                    parameters=[
                        # Use all parameters from handler
                        *inspect.signature(f).parameters.values(),
                        inspect.Parameter(
                            name="_request",
                            kind=inspect.Parameter.VAR_POSITIONAL,
                            annotation=Request,
                        ),
                    ],
                )
            else:
                wrapper = f

            return wrapper

        self.access_wrapper = (
            multisite_access_wrapper if self.multi_site_enabled else access_wrapper
        )

    def choose_router(self, f):
        if peewee_enabled or getattr(f, "skip_dbtransaction", False):
            if getattr(f, "auth_by_cookie_or_header", False):
                if getattr(f, "login_optional", False):
                    return self.optional_auth_by_cookie_or_header_router
                else:  # login_required
                    f.login_required = True
                    return self.auth_by_cookie_or_header_router

            elif getattr(f, "login_optional", False):
                return self.optional_auth_by_header_router

            elif getattr(f, "skip_authorization", False):
                return self.unsecure_router

            else:  # login_required
                f.login_required = True
                return self.auth_by_header_router

        # For piccolo, dbtransaction is handled by separate routers
        elif getattr(f, "auth_by_cookie_or_header", False):
            if getattr(f, "login_optional", False):
                return self.optional_auth_by_cookie_or_header_router_with_dbtransaction
            else:  # login_required
                f.login_required = True
                return self.auth_by_cookie_or_header_router_with_dbtransaction

        elif getattr(f, "login_optional", False):
            return self.optional_auth_by_header_router_with_dbtransaction

        elif getattr(f, "skip_authorization", False):
            return self.unsecure_router_with_dbtransaction

        else:  # login_required
            f.login_required = True
            return self.auth_by_header_router_with_dbtransaction

    def build(self, method, method_args, method_kw, f):
        module = f.__module__.split(".")[-1].strip("_")
        name = f.__name__.strip("_")
        response_model = getattr(f, "response_model", None)
        db_tr_wrapper = (
            phony
            if peewee_enabled is False or getattr(f, "skip_dbtransaction", False)
            else self.db_tr_wrapper
        )

        if "operation_id" not in method_kw:
            method_kw["operation_id"] = f"{name}_{module}"
        if "name" not in method_kw:
            method_kw["name"] = method_kw["operation_id"]
        if "tags" not in method_kw:
            method_kw["tags"] = [module]

        if response_model is not None and "response_model" not in method_kw:
            method_kw["response_model"] = response_model

        if (
            "response_model" in method_kw
            and "response_model_exclude_unset" not in method_kw
        ):
            method_kw["response_model_exclude_unset"] = True

        print(
            f"{method_args[0]}",
            f"[{method.__name__.upper()}] => {f.__module__}:{f.__name__}",
        )
        m = method(*method_args, **method_kw)
        f = self.access_wrapper(
            self.honeybadger_wrapper(raise_not_found_on_none(db_tr_wrapper(f)))
        )
        # NOTE: ^ wrapper ordering is important. access_wrapper needs request which
        # others don't. If access_wrapper comes late in the order it won't be passed
        # request parameter.
        return m(f)

    def get(self, path, *a, **k):
        def _wrapper(f):
            router = self.choose_router(f)
            full_path = (
                path if path.startswith("/") else f"{self.urls_prefix}/{path}"
            ).rstrip("/")
            args = (full_path,) + a
            return self.build(router.get, args, k, f)

        return _wrapper

    def post(self, path, *a, **k):
        def _wrapper(f):
            router = self.choose_router(f)
            full_path = (
                path if path.startswith("/") else f"{self.urls_prefix}/{path}"
            ).rstrip("/")
            args = (full_path,) + a
            return self.build(router.post, args, k, f)

        return _wrapper

    def put(self, path, *a, **k):
        def _wrapper(f):
            router = self.choose_router(f)
            full_path = (
                path if path.startswith("/") else f"{self.urls_prefix}/{path}"
            ).rstrip("/")
            args = (full_path,) + a
            return self.build(router.put, args, k, f)

        return _wrapper

    def patch(self, path, *a, **k):
        def _wrapper(f):
            router = self.choose_router(f)
            full_path = (
                path if path.startswith("/") else f"{self.urls_prefix}/{path}"
            ).rstrip("/")
            args = (full_path,) + a
            return self.build(router.patch, args, k, f)

        return _wrapper

    def delete(self, path, *a, **k):
        def _wrapper(f):
            router = self.choose_router(f)
            full_path = (
                path if path.startswith("/") else f"{self.urls_prefix}/{path}"
            ).rstrip("/")
            args = (full_path,) + a
            return self.build(router.delete, args, k, f)

        return _wrapper

    def map_resource(self, collection_url, resource=None, handlers=None, id_field="id"):
        if resource:
            raise NotImplementedError("Resource not supported yet")

        collection_url = collection_url.rstrip("/")
        resource_url = f"{collection_url}/{{{id_field}}}"
        assert isinstance(handlers, (list, tuple)), "handlers should be list or tuple"
        (
            get_collection,
            add_resource,
            replace_resource,
            get_resource,
            update_resource,
            delete_resource,
        ) = handlers

        if get_collection:
            self.get(collection_url)(get_collection)
        if add_resource:
            self.post(collection_url)(add_resource)
        if replace_resource:
            self.put(resource_url)(replace_resource)
        if get_resource:
            self.get(resource_url)(get_resource)
        if update_resource:
            self.patch(resource_url)(update_resource)
        if delete_resource:
            self.delete(resource_url)(delete_resource)


@ep.login_required
def whoami(user: User = AuthParams.user):
    return user.to_dict()
